'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _slicedToArray = function () { function sliceIterator(arr, i) { var _arr = []; var _n = true; var _d = false; var _e = undefined; try { for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) { _arr.push(_s.value); if (i && _arr.length === i) break; } } catch (err) { _d = true; _e = err; } finally { try { if (!_n && _i["return"]) _i["return"](); } finally { if (_d) throw _e; } } return _arr; } return function (arr, i) { if (Array.isArray(arr)) { return arr; } else if (Symbol.iterator in Object(arr)) { return sliceIterator(arr, i); } else { throw new TypeError("Invalid attempt to destructure non-iterable instance"); } }; }();

exports.createJobs = createJobs;

var _fs = require('fs');

var _fs2 = _interopRequireDefault(_fs);

var _path = require('path');

var _path2 = _interopRequireDefault(_path);

var _imurmurhash = require('imurmurhash');

var _imurmurhash2 = _interopRequireDefault(_imurmurhash);

var _mimeTypes = require('mime-types');

var mimeTypes = _interopRequireWildcard(_mimeTypes);

var _babelCore = require('babel-core');

var babel = _interopRequireWildcard(_babelCore);

var _babylon = require('babylon');

var babylon = _interopRequireWildcard(_babylon);

var _postcss = require('postcss');

var _postcss2 = _interopRequireDefault(_postcss);

var _builtins = require('browserify/lib/builtins');

var _builtins2 = _interopRequireDefault(_builtins);

var _browserResolve2 = require('browser-resolve');

var _browserResolve3 = _interopRequireDefault(_browserResolve2);

var _promisifyNode = require('promisify-node');

var _promisifyNode2 = _interopRequireDefault(_promisifyNode);

var _string = require('lodash/string');

var _collection = require('lodash/collection');

var _array = require('lodash/array');

var _object = require('lodash/object');

var _lang = require('lodash/lang');

var _babylonAstDependencies = require('babylon-ast-dependencies');

var _babylonAstDependencies2 = _interopRequireDefault(_babylonAstDependencies);

var _postcssAstDependencies = require('postcss-ast-dependencies');

var _postcssAstDependencies2 = _interopRequireDefault(_postcssAstDependencies);

var _babelGenerator = require('babel-generator');

var _babelGenerator2 = _interopRequireDefault(_babelGenerator);

var _utils = require('./utils');

function _interopRequireWildcard(obj) { if (obj && obj.__esModule) { return obj; } else { var newObj = {}; if (obj != null) { for (var key in obj) { if (Object.prototype.hasOwnProperty.call(obj, key)) newObj[key] = obj[key]; } } newObj.default = obj; return newObj; } }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const readFile = (0, _promisifyNode2.default)(_fs2.default.readFile);
const stat = (0, _promisifyNode2.default)(_fs2.default.stat);
const browserResolve = (0, _promisifyNode2.default)(_browserResolve3.default);

function createJobs() {
  var _ref = arguments.length <= 0 || arguments[0] === undefined ? {} : arguments[0];

  let getState = _ref.getState;

  return {
    ready: function (ref, store) {
      // All the jobs that must be completed before
      // the record is emitted
      return Promise.all([ref.name, store.hash(ref), store.content(ref), store.moduleDefinition(ref), store.url(ref), store.sourceMapAnnotation(ref), store.hashedFilename(ref), store.isTextFile(ref), store.mimeType(ref), store.fileDependencies(ref)]);
    },
    basename: function (ref) {
      return Promise.resolve(_path2.default.basename(ref.name, _path2.default.extname(ref.name)));
    },
    ext: function (ref) {
      return Promise.resolve(_path2.default.extname(ref.name));
    },
    isTextFile: function (ref, store) {
      return store.ext(ref).then(ext => {
        return ext === '.js' || ext === '.css' || ext === '.json';
      });
    },
    mimeType: function (ref, store) {
      return store.ext(ref).then(ext => mimeTypes.lookup(ext) || null);
    },
    readText: function (ref) {
      return readFile(ref.name, 'utf8');
    },
    stat: function (ref) {
      return stat(ref.name);
    },
    mtime: function (ref, store) {
      return store.stat(ref).then(stat => {
        return stat.mtime.getTime();
      });
    },
    hashText: function (ref, store) {
      return store.readText(ref).then(text => {
        const hash = new _imurmurhash2.default(text).result();
        return hash.toString();
      });
    },
    hash: function (ref, store) {
      return store.isTextFile(ref).then(isTextFile => {
        if (isTextFile) {
          return store.hashText(ref);
        } else {
          return store.mtime(ref);
        }
      })
      // We coerce everything to a string for consistency
      .then(hash => hash.toString());
    },
    hashedFilename: function (ref, store) {
      return Promise.all([store.basename(ref), store.hash(ref), store.ext(ref)]).then(_ref2 => {
        var _ref3 = _slicedToArray(_ref2, 3);

        let basename = _ref3[0];
        let hash = _ref3[1];
        let ext = _ref3[2];

        return `${ basename }-${ hash }${ ext }`;
      });
    },
    hashedName: function (ref, store) {
      return store.hashedFilename(ref).then(hashedFilename => {
        return _path2.default.join(_path2.default.dirname(ref.name), hashedFilename);
      });
    },
    cache: function () {
      return Promise.resolve(getState().jobCache);
    },
    cacheKey: function (ref, store) {
      return store.isTextFile(ref).then(isTextFile => {
        const key = [ref.name, store.mtime(ref)];

        if (isTextFile) {
          key.push(store.hash(ref));
        }

        return Promise.all(key);
      });
    },
    readCache: function (ref, store) {
      return Promise.all([store.cache(ref), store.cacheKey(ref)]).then(_ref4 => {
        var _ref5 = _slicedToArray(_ref4, 2);

        let cache = _ref5[0];
        let key = _ref5[1];
        return cache.get(key);
      }).then(data => {
        if ((0, _lang.isNull)(data)) {
          return {};
        }
        return data;
      });
    },
    writeCache: function (ref, store) {
      return Promise.all([store.cache(ref), store.cacheKey(ref), store.readCache(ref)]).then(_ref6 => {
        var _ref7 = _slicedToArray(_ref6, 3);

        let cache = _ref7[0];
        let key = _ref7[1];
        let cacheData = _ref7[2];

        return cache.set(key, cacheData);
      });
    },
    url: function (ref, store) {
      return store.isTextFile(ref).then(isTextFile => {
        if (isTextFile) {
          return store.hashedName(ref);
        } else {
          return ref.name;
        }
      }).then(name => {
        var _getState = getState();

        const sourceRoot = _getState.sourceRoot;
        const rootUrl = _getState.rootUrl;

        // Try to produce a more readable url, but fallback to an absolute
        // path if necessary. The fallback is necessary if the record is
        // being pulled in from outside the source root, eg: by a symbolic
        // link

        let relPath;
        if ((0, _string.startsWith)(name, sourceRoot)) {
          relPath = _path2.default.relative(sourceRoot, name);
        } else {
          relPath = name;
        }

        return rootUrl + relPath.split(_path2.default.ext).join('/');
      });
    },

    /**
     * A url to the original content of the record
     */
    sourceUrl: function (ref, store) {
      return store.hash(ref).then(hash => {
        // We append the hash to get around browser caching of
        // source maps
        return 'file://' + ref.name + '?' + hash;
      });
    },
    sourceMapAnnotation: function (ref, store) {
      return Promise.all([store.ext(ref), store.sourceMap(ref)]).then(_ref8 => {
        var _ref9 = _slicedToArray(_ref8, 2);

        let ext = _ref9[0];
        let sourceMap = _ref9[1];

        if (!sourceMap || !(0, _collection.includes)(['.css', '.js', '.json'], ext)) {
          return null;
        }

        const base64SourceMap = new Buffer(sourceMap).toString('base64');
        const body = 'sourceMappingURL=data:application/json;charset=utf-8;base64,' + base64SourceMap;

        if (ext === '.css') {
          return `\n/*# ${ body } */`;
        } else {
          return '\n//# ' + body;
        }
      });
    },
    postcssPlugins: function () {
      return Promise.resolve([]);
    },
    postcssTransformOptions: function (ref, store) {
      return Promise.all([store.url(ref), store.sourceUrl(ref)]).then(_ref10 => {
        var _ref11 = _slicedToArray(_ref10, 2);

        let url = _ref11[0];
        let sourceUrl = _ref11[1];

        return {
          from: sourceUrl,
          to: url,
          // Generate a source map, but keep it separate from the code
          map: {
            inline: false,
            annotation: false
          }
        };
      });
    },
    postcssTransform: function (ref, store) {
      return Promise.all([store.readText(ref), store.postcssPlugins(ref), store.postcssTransformOptions(ref)]).then(_ref12 => {
        var _ref13 = _slicedToArray(_ref12, 3);

        let text = _ref13[0];
        let plugins = _ref13[1];
        let options = _ref13[2];


        // Finds any `@import ...` and `url(...)` identifiers and
        // annotates the result object
        const analyzeDependencies = _postcss2.default.plugin('unfort-analyze-dependencies', () => {
          return (root, result) => {
            result.unfortDependencies = (0, _postcssAstDependencies2.default)(root);
          };
        });

        // As we serve the files with different names, we need to remove
        // the `@import ...` rules
        const removeImports = _postcss2.default.plugin('unfort-remove-imports', () => {
          return root => {
            root.walkAtRules('import', rule => rule.remove());
          };
        });

        plugins = plugins.concat([analyzeDependencies, removeImports]);

        return (0, _postcss2.default)(plugins).process(text, options);
      });
    },
    babelTransformOptions: function (ref, store) {
      return Promise.all([store.url(ref), store.sourceUrl(ref)]).then(_ref14 => {
        var _ref15 = _slicedToArray(_ref14, 2);

        let url = _ref15[0];
        let sourceUrl = _ref15[1];

        return {
          filename: ref.name,
          sourceType: 'module',
          sourceMaps: true,
          sourceMapTarget: url,
          sourceFileName: sourceUrl
        };
      });
    },
    babelTransform: function (ref, store) {
      return Promise.all([store.readText(ref), store.babelTransformOptions(ref)]).then(_ref16 => {
        var _ref17 = _slicedToArray(_ref16, 2);

        let text = _ref17[0];
        let options = _ref17[1];

        return babel.transform(text, options);
      });
    },
    babelGeneratorOptions: function (ref, store) {
      return Promise.all([store.url(ref), store.sourceUrl(ref)]).then(_ref18 => {
        var _ref19 = _slicedToArray(_ref18, 2);

        let url = _ref19[0];
        let sourceUrl = _ref19[1];

        // We want to preserve the compression applied to vendor assets
        const shouldMinify = (0, _string.startsWith)(ref.name, getState().vendorRoot);

        return {
          sourceMaps: true,
          sourceMapTarget: url,
          sourceFileName: sourceUrl,
          minified: shouldMinify
        };
      });
    },
    babelGenerator: function (ref, store) {
      return Promise.all([store.readText(ref), store.babylonAst(ref), store.babelGeneratorOptions(ref)]).then(_ref20 => {
        var _ref21 = _slicedToArray(_ref20, 3);

        let text = _ref21[0];
        let ast = _ref21[1];
        let options = _ref21[2];

        return (0, _babelGenerator2.default)(ast, options, text);
      });
    },
    shouldBabelTransform: function (ref) {
      var _getState2 = getState();

      const rootNodeModules = _getState2.rootNodeModules;
      const vendorRoot = _getState2.vendorRoot;


      return Promise.resolve(!(0, _string.startsWith)(ref.name, rootNodeModules) && !(0, _string.startsWith)(ref.name, vendorRoot));
    },
    babelFile: function (ref, store) {
      return store.shouldBabelTransform(ref).then(shouldBabelTransform => {
        if (shouldBabelTransform) {
          return store.babelTransform(ref);
        } else {
          return store.babelGenerator(ref);
        }
      });
    },
    babelAst: function (ref, store) {
      return store.babelTransform(ref).then(file => file.ast);
    },
    babylonAst: function (ref, store) {
      return store.readText(ref).then(text => {
        return babylon.parse(text, {
          sourceType: 'script'
        });
      });
    },
    ast: function (ref, store) {
      return store.ext(ref).then(ext => {
        if (ext === '.js') {
          return store.shouldBabelTransform(ref).then(shouldBabelTransform => {
            if (shouldBabelTransform) {
              return store.babelAst(ref);
            } else {
              return store.babylonAst(ref);
            }
          });
        }

        // Note: we reject the `ast` job for .css files as we handle
        // it during the initial traversal and transformation in the
        // `postcssTransform` job
        throw new Error(`Unknown extension "${ ext }", cannot parse "${ ref.name }"`);
      });
    },
    analyzeDependencies: function (ref, store) {
      return store.ext(ref).then(ext => {
        if (ext === '.css') {
          return store.postcssTransform(ref).then(result => {
            return result.unfortDependencies;
          });
        }

        if (ext === '.js') {
          return store.ast(ref).then(ast => (0, _babylonAstDependencies2.default)(ast));
        }

        return [];
      });
    },
    dependencyIdentifiers: function (ref, store) {
      return store.readCache(ref).then(cachedData => {
        if (cachedData.dependencyIdentifiers) {
          return cachedData.dependencyIdentifiers;
        }

        return store.analyzeDependencies(ref).then(deps => deps.map(dep => dep.source))
        // Remove any parts of the identifier that we wont be able to
        // map to the file system
        .then(ids => ids.map(id => {
          // Webpack loaders
          const bangStart = id.indexOf('!');
          if (bangStart !== -1) {
            id = id.slice(0, bangStart);
          }

          // Url params
          const paramStart = id.indexOf('?');
          if (paramStart !== -1) {
            id = id.slice(0, paramStart);
          }

          // Url hashes
          const hashStart = id.indexOf('#');
          if (hashStart !== -1) {
            id = id.slice(0, hashStart);
          }

          return id;
        })).then(ids => cachedData.dependencyIdentifiers = ids);
      });
    },
    pathDependencyIdentifiers: function (ref, store) {
      return store.dependencyIdentifiers(ref).then(ids => ids.filter(id => id[0] === '.' || _path2.default.isAbsolute(id)));
    },
    packageDependencyIdentifiers: function (ref, store) {
      return store.dependencyIdentifiers(ref).then(ids => ids.filter(id => id[0] !== '.' && !_path2.default.isAbsolute(id)));
    },
    resolver: function (ref, store) {
      return store.resolverOptions(ref).then(options => {
        // We use `browser-resolve` to resolve ids as it picks up browser-specific
        // entry points for packages
        return id => browserResolve(id, options);
      });
    },
    resolverOptions: function (ref) {
      return Promise.resolve({
        // The directory that the resolver starts in when looking for a file
        // to matches an identifier
        basedir: _path2.default.dirname(ref.name),
        // The extensions that the resolver looks for considering identifiers
        // without an extension
        extensions: ['.js', '.json'],
        // The node core modules that should be shimmed for browser environments.
        // We use browserify's as they tend to upgrade them more often. Webpack's
        // `node-libs-browser` is another alternative
        modules: _builtins2.default
      });
    },

    /**
     * If a dependency identifier is relative (./ ../) or absolute (/), there are
     * edge-cases where caching the resolved path may produce the wrong result.
     * For example: an identifier "./foo" may resolve to either a "./foo.js" or
     * or "./foo/index.js". Detecting these cases is problematic, so we avoid the
     * problem by ensuring that the resolver always inspects the file system for
     * path-based identifiers originating from files that we expect to change
     * frequently.
     *
     * The one exception is files that live in node_modules, we aggressively cache
     * these as we assume they are static. This enables us to avoid some costly IO
     * overhead when possible
     */
    shouldCacheResolvedPathDependencies: function (ref) {
      return Promise.resolve((0, _string.startsWith)(ref.name, getState().rootNodeModules));
    },

    /**
     * If a dependency identifier refers to a package (eg: is not a path-based
     * identifier), we can cache the resolved path and leave higher levels to
     * perform cache invalidation
     */
    shouldCacheResolvedPackageDependencies: function () {
      return Promise.resolve(true);
    },
    resolvePathDependencies: function (ref, store) {
      return Promise.all([store.readCache(ref), store.shouldCacheResolvedPathDependencies(ref)]).then(_ref22 => {
        var _ref23 = _slicedToArray(_ref22, 2);

        let cachedData = _ref23[0];
        let shouldCache = _ref23[1];

        if (shouldCache && cachedData.resolvePathDependencies) {
          return cachedData.resolvePathDependencies;
        }

        return store.pathDependencyIdentifiers(ref).then(ids => store.resolver(ref).then(resolver => Promise.all(ids.map(id => resolver(id)))).then(resolved => (0, _array.zipObject)(ids, resolved))).then(deps => cachedData.resolvePathDependencies = deps);
      });
    },
    resolvePackageDependencies: function (ref, store) {
      return Promise.all([store.readCache(ref), store.shouldCacheResolvedPackageDependencies(ref)]).then(_ref24 => {
        var _ref25 = _slicedToArray(_ref24, 2);

        let cachedData = _ref25[0];
        let shouldCache = _ref25[1];

        if (shouldCache && cachedData.resolvePackageDependencies) {
          return cachedData.resolvePackageDependencies;
        }

        return store.packageDependencyIdentifiers(ref).then(ids => store.resolver(ref).then(resolver => Promise.all(ids.map(id => resolver(id)))).then(resolved => (0, _array.zipObject)(ids, resolved))).then(deps => cachedData.resolvePackageDependencies = deps);
      });
    },
    resolvedDependencies: function (ref, store) {
      return Promise.all([store.resolvePackageDependencies(ref), store.resolvePathDependencies(ref)]).then(_ref26 => {
        var _ref27 = _slicedToArray(_ref26, 2);

        let pathDeps = _ref27[0];
        let packageDeps = _ref27[1];

        return (0, _object.assign)({}, pathDeps, packageDeps);
      });
    },
    code: function (ref, store) {
      return store.isTextFile(ref).then(isTextFile => {
        if (!isTextFile) {
          return null;
        }

        return Promise.all([store.ext(ref), store.readCache(ref)]).then(_ref28 => {
          var _ref29 = _slicedToArray(_ref28, 2);

          let ext = _ref29[0];
          let cachedData = _ref29[1];

          if (cachedData.code) {
            return cachedData.code;
          }

          if (ext === '.css') {
            return store.postcssTransform(ref).then(result => cachedData.code = result.css);
          }

          if (ext === '.js') {
            const state = getState();
            // Serve up the runtimes without any transformation
            if (ref.name === state.bootstrapRuntime || ref.name === state.hotRuntime) {
              return store.readText(ref);
            }

            return store.babelFile(ref).then(file => cachedData.code = file.code);
          }

          if (ext === '.json') {
            return store.readText(ref);
          }

          return Promise.reject(`Unknown text file extension: ${ ext }. Cannot generate code for file: ${ ref.name }`);
        });
      });
    },

    /**
     * For JS and JSON records, we can inject the record's code. For all other
     * types, we inject a url to their location
     */
    moduleContents: function (ref, store) {
      return store.ext(ref).then(ext => {
        if (ext === '.js' || ext === '.json') {
          return store.code(ref);
        } else {
          return store.url(ref).then(url => JSON.stringify(url));
        }
      });
    },

    /**
     * Indicates if the record requires a shim module to be defined. These shim
     * modules are used so that the runtime can interact with a representation
     * of non-JS records
     */
    shouldShimModuleDefinition: function (ref, store) {
      return store.ext(ref).then(ext => ext !== '.js');
    },

    /**
     * Generates the module code for a record. This is primarily of use to
     * create shim modules for non-js records
     */
    moduleCode: function (ref, store) {
      return Promise.all([store.shouldShimModuleDefinition(ref), store.moduleContents(ref)]).then(_ref30 => {
        var _ref31 = _slicedToArray(_ref30, 2);

        let shouldShimModuleDefinition = _ref31[0];
        let moduleContents = _ref31[1];

        if (!shouldShimModuleDefinition) {
          return moduleContents;
        }

        // We fake babel's ES => commonjs shim so that the hot runtime knows
        // that `module.exports` will never be a function and hence a proxy
        // object can be used
        return ['Object.defineProperty(exports, "__esModule", {', '  value: true', '});', `exports["default"] = ${ moduleContents };`, 'if (module.hot) {', '  module.hot.accept();', '}'].join('\n');
      });
    },

    /**
     * Create the module definition that we use to inject a record into
     * the runtime
     */
    moduleDefinition: function (ref, store) {
      // The bootstrap is the one file that we need to ensure is pushed
      // to the client without any shims or wrappers
      if (ref.name === getState().bootstrapRuntime) {
        return Promise.resolve(null);
      }

      return Promise.all([store.resolvedDependencies(ref), store.hash(ref), store.moduleCode(ref)]).then(_ref32 => {
        var _ref33 = _slicedToArray(_ref32, 3);

        let resolvedDependencies = _ref33[0];
        let hash = _ref33[1];
        let moduleCode = _ref33[2];

        return (0, _utils.createJSModuleDefinition)({
          name: ref.name,
          deps: resolvedDependencies,
          hash: hash,
          code: moduleCode
        });
      });
    },

    /**
     * Generates the executable content of a record.
     * For js and json records, this is the module definition.
     * For css, this is the stylesheet.
     */
    content: function (ref, store) {
      return Promise.all([store.isTextFile(ref), store.ext(ref)]).then(_ref34 => {
        var _ref35 = _slicedToArray(_ref34, 2);

        let isTextFile = _ref35[0];
        let ext = _ref35[1];

        if (!isTextFile) {
          return null;
        }

        if (ref.name === getState().bootstrapRuntime || ext === '.css') {
          return store.code(ref);
        }

        if (ext === '.js' || ext === '.json') {
          return store.moduleDefinition(ref);
        }

        return Promise.reject(`Unknown text file extension: ${ ext }. Cannot generate content for file: ${ ref.name }`);
      });
    },

    /**
     * Generates a textual representation of a record's source map
     */
    sourceMap: function (ref, store) {
      return store.isTextFile(ref).then(isTextFile => {
        if (!isTextFile) {
          return null;
        }

        return store.readCache(ref).then(cachedData => {
          if (cachedData.sourceMap) {
            return cachedData.sourceMap;
          }

          return store.ext(ref).then(ext => {
            if (ext === '.css') {
              return store.postcssTransform(ref).then(result => cachedData.sourceMap = result.map.toString());
            }

            if (ext === '.js') {
              return store.babelFile(ref).then(file => {
                // Offset each line in the source map to reflect the call to
                // the module runtime
                file.map.mappings = _utils.JS_MODULE_SOURCE_MAP_LINE_OFFSET + file.map.mappings;

                return cachedData.sourceMap = JSON.stringify(file.map);
              });
            }

            if (ext === '.json') {
              return null;
            }

            return Promise.reject(`Unknown text file extension: ${ ext }. Cannot generate source map for file: ${ ref.name }`);
          });
        });
      });
    },

    /**
     * If any other files were used to generate this record, this should
     * this should return an array of paths to those files. This enables
     * the file watchers to invalidate a record when its dependencies
     * change.
     *
     * An example use-case would be compiling a CSS file with LESS or
     * SASS. As those tools will bundle multiple files into a single file,
     * you should return an array of the bundled files
     */
    fileDependencies: function () {
      return Promise.resolve([]);
    }
  };
}
//# sourceMappingURL=jobs.js.map