'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.State = exports.vendorRoot = exports.hotRuntime = exports.bootstrapRuntime = undefined;
exports.createState = createState;

var _path = require('path');

var _path2 = _interopRequireDefault(_path);

var _immutable = require('immutable');

var _immutable2 = _interopRequireDefault(_immutable);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const bootstrapRuntime = exports.bootstrapRuntime = require.resolve('../runtimes/bootstrap');
const hotRuntime = exports.hotRuntime = require.resolve('../runtimes/hot');
const vendorRoot = exports.vendorRoot = _path2.default.normalize(_path2.default.join(__dirname, '..', 'vendor'));

/**
 * The `state` object passed around, most commonly via `getState`.
 *
 * An immutable record that contains a mixture of configuration, generated
 * data, and references to stateful sub-modules
 */
const State = exports.State = _immutable2.default.Record({
  // ===================
  // Build configuration
  // ===================

  // The dependency graph's entry points
  entryPoints: null,
  // The root of your project
  sourceRoot: null,
  // The root `node_modules` directory that your modules are pulled from
  rootNodeModules: null,
  // The root directory that cached data will be dumped into
  cacheDirectory: null,
  // The record store's jobs
  jobs: null,
  // The directory containing the hot runtime's vendor packages
  vendorRoot: vendorRoot,
  rootUrl: null,
  getSockets: null,

  // ================
  // Environment hash
  // ================

  // Options passed to `envHash`
  envHash: null,
  // The hash generated by `envHash`
  environmentHash: null,

  // =================
  // The module system
  // =================

  bootstrapRuntime: bootstrapRuntime,
  hotRuntime: hotRuntime,

  // ===================
  // Stateful subsystems
  // ===================

  recordStore: null,
  jobCache: null,
  graph: null,
  watchers: null,

  // =======
  // Loggers
  // =======

  // We use wrappers around the console object so that we can prevent
  // output during tests
  logInfo: function () {
    console.info.apply(console, arguments);
  },
  logError: function () {
    console.error.apply(console, arguments);
  },

  // =============================
  // Data generated from the build
  // =============================

  // Any errors that have blocked a build
  errors: null,
  // The latest snapshot of the dependency graph
  nodes: null,
  // The latest snapshot of the record store
  records: null,
  // Used to map the `url` job to a specific record
  recordsByUrl: null,
  // Used to map the `fileDependencies` job to one or more records
  recordsByFileDependencies: null
});

/**
 * Creates a state object and sets defaults for required values
 *
 * @param {object} overrides
 */
function createState() {
  let overrides = arguments.length <= 0 || arguments[0] === undefined ? {} : arguments[0];

  let state = State(overrides);

  if (!state.getSockets) {
    state = state.set('getSockets', () => []);
  }

  if (!state.sourceRoot) {
    state = state.set('sourceRoot', process.cwd());
  }

  if (!state.rootNodeModules) {
    state = state.set('rootNodeModules', _path2.default.join(state.sourceRoot, 'node_modules'));
  }

  if (!state.cacheDirectory) {
    state = state.set('cacheDirectory', _path2.default.join(state.sourceRoot, '.unfort'));
  }

  if (!state.entryPoints) {
    state = state.set('entryPoints', []);
  }

  return state;
}
//# sourceMappingURL=state.js.map